@@ -75,6 +75,8 @@ static int incremental_mode = 0;
 static int partition_cnt = 0;
 static int eof_cnt = 0;
 static int with_dr = 1;
+static int read_hdrs = 0;
+
 
 static void stop (int sig) {
         if (!run)
@@ -315,6 +317,12 @@ static void msg_consume (rd_kafka_message_t *rkmessage, void *opaque) {
 
         }
 
+        if (read_hdrs) {
+                rd_kafka_headers_t *hdrs;
+                /* Force parsing of headers but don't do anything with them. */
+                rd_kafka_message_headers(rkmessage, &hdrs);
+        }
+
         if (msgcnt != -1 && (int)cnt.msgs >= msgcnt)
                 run = 0;
 }
@@ -715,6 +723,61 @@ static int read_conf_file (rd_kafka_conf_t *conf,
 }
 
 
+static rd_kafka_resp_err_t do_produce (rd_kafka_t *rk,
+                                       rd_kafka_topic_t *rkt, int32_t partition,
+                                       int msgflags,
+                                       void *payload, size_t size,
+                                       const void *key, size_t key_size,
+                                       const rd_kafka_headers_t *hdrs) {
+
+        /* Send/Produce message. */
+        if (hdrs) {
+                rd_kafka_headers_t *hdrs_copy;
+                rd_kafka_resp_err_t err;
+
+                hdrs_copy = rd_kafka_headers_copy(hdrs);
+
+                err = rd_kafka_producev(
+                        rk,
+                        RD_KAFKA_V_RKT(rkt),
+                        RD_KAFKA_V_PARTITION(partition),
+                        RD_KAFKA_V_MSGFLAGS(msgflags),
+                        RD_KAFKA_V_VALUE(payload, size),
+                        RD_KAFKA_V_KEY(key, key_size),
+                        RD_KAFKA_V_HEADERS(hdrs_copy),
+                        RD_KAFKA_V_END);
+
+                if (err)
+                        rd_kafka_headers_destroy(hdrs_copy);
+
+                return err;
+
+        } else {
+                if (rd_kafka_produce(rkt, partition, msgflags, payload, size,
+                                     key, key_size, NULL) == -1)
+                        return rd_kafka_last_error();
+        }
+
+        return RD_KAFKA_RESP_ERR_NO_ERROR;
+}
+
+/**
+ * @brief Sleep for \p sleep_us microseconds.
+ */
+static void do_sleep (int sleep_us) {
+        if (sleep_us > 100) {
+#ifdef _MSC_VER
+                Sleep(sleep_us / 1000);
+#else
+                usleep(sleep_us);
+#endif
+        } else {
+                rd_ts_t next = rd_clock() + (rd_ts_t)sleep_us;
+                while (next > rd_clock())
+                        ;
+        }
+}
+
 
 int main (int argc, char **argv) {
 	char *brokers = NULL;
@@ -749,6 +812,8 @@ int main (int argc, char **argv) {
         int rate_sleep = 0;
 	rd_kafka_topic_partition_list_t *topics;
         int exitcode = 0;
+        rd_kafka_headers_t *hdrs = NULL;
+        rd_kafka_resp_err_t err;
 
 	/* Kafka configuration */
 	conf = rd_kafka_conf_new();
@@ -789,7 +854,7 @@ int main (int argc, char **argv) {
 	while ((opt =
 		getopt(argc, argv,
 		       "PCG:t:p:b:s:k:c:fi:MDd:m:S:x:"
-                       "R:a:z:o:X:B:eT:Y:qvIur:lA:OwN")) != -1) {
+                       "R:a:z:o:X:B:eT:Y:qvIur:lA:OwNHH:")) != -1) {
 		switch (opt) {
 		case 'G':
 			if (rd_kafka_conf_set(conf, "group.id", optarg,
@@ -888,6 +953,37 @@ int main (int argc, char **argv) {
 		case 'd':
 			debug = optarg;
 			break;
+                case 'H':
+                {
+                        char *name, *val;
+                        size_t name_sz = -1;
+
+                        if (!optarg) {
+                                read_hdrs = 1;
+                                break;
+                        }
+
+                        name = optarg;
+                        val = strchr(name, '=');
+                        if (val) {
+                                name_sz = (size_t)(val-name);
+                                val++; /* past the '=' */
+                        }
+
+                        if (!hdrs)
+                                hdrs = rd_kafka_headers_new(8);
+
+                        err = rd_kafka_header_add(hdrs, name, name_sz, val, -1);
+                        if (err) {
+                                fprintf(stderr,
+                                        "%% Failed to add header %s: %s\n",
+                                        name, rd_kafka_err2str(err));
+                                exit(1);
+                        }
+
+                        read_hdrs = 1;
+                }
+                break;
 		case 'X':
 		{
 			char *name, *val;
@@ -1033,6 +1129,8 @@ int main (int argc, char **argv) {
 			"  -b <brokers> Broker address list (host[:port],..)\n"
 			"  -s <size>    Message size (producer)\n"
 			"  -k <key>     Message key (producer)\n"
+                        "  -H <name[=value]> Add header to message (producer)\n"
+                        "  -H           Read message headers (consumer)\n"
 			"  -c <cnt>     Messages to transmit/receive\n"
 			"  -x <cnt>     Hard exit after transmitting <cnt> messages (producer)\n"
 			"  -D           Copy/Duplicate data buffer (producer)\n"
@@ -1261,10 +1359,9 @@ int main (int argc, char **argv) {
 
 			cnt.tx++;
 			while (run &&
-			       rd_kafka_produce(rkt, partition,
-						sendflags, pbuf, msgsize,
-						key, keylen, NULL) == -1) {
-				rd_kafka_resp_err_t err = rd_kafka_last_error();
+                               (err = do_produce(rk, rkt, partition, sendflags,
+                                                 pbuf, msgsize,
+                                                 key, keylen, hdrs))) {
 				if (err == RD_KAFKA_RESP_ERR__UNKNOWN_PARTITION)
 					printf("%% No such partition: "
 						   "%"PRId32"\n", partition);
@@ -1303,19 +1400,8 @@ int main (int argc, char **argv) {
 			cnt.msgs++;
 			cnt.bytes += msgsize;
 
-                        if (rate_sleep) {
-				if (rate_sleep > 100) {
-#ifdef _MSC_VER
-					Sleep(rate_sleep / 1000);
-#else
-					usleep(rate_sleep);
-#endif
-				} else {
-					rd_ts_t next = rd_clock() + rate_sleep;
-					while (next > rd_clock())
-						;
-				}
-                        }
+                        if (rate_sleep)
+                                do_sleep(rate_sleep);
 
 			/* Must poll to handle delivery reports */
 			rd_kafka_poll(rk, 0);
@@ -1443,6 +1529,12 @@ int main (int argc, char **argv) {
                         if (r == -1)
                                 fprintf(stderr, "%% Error: %s\n",
                                         rd_kafka_err2str(rd_kafka_last_error()));
+                        else if (r > 0 && rate_sleep) {
+                                /* Simulate processing time
+                                 * if `-r <rate>` was set. */
+                                do_sleep(rate_sleep);
+                        }
+
 
 			print_stats(rk, mode, otype, compression);
 
@@ -1524,6 +1616,11 @@ int main (int argc, char **argv) {
 			if (rkmessage) {
 				msg_consume(rkmessage, NULL);
 				rd_kafka_message_destroy(rkmessage);
+
+                                /* Simulate processing time
+                                 * if `-r <rate>` was set. */
+                                if (rate_sleep)
+                                        do_sleep(rate_sleep);
 			}
 
 			cnt.t_fetch_latency += rd_clock() - fetch_latency;
@@ -1540,6 +1637,9 @@ int main (int argc, char **argv) {
 		rd_kafka_destroy(rk);
 	}
 
+        if (hdrs)
+                rd_kafka_headers_destroy(hdrs);
+
 	print_stats(NULL, mode, otype|_OTYPE_FORCE, compression);
 
 	if (cnt.t_fetch_latency && cnt.msgs)
