@@ -26,7 +26,8 @@
  * POSSIBILITY OF SUCH DAMAGE.
  */
 
-#pragma once
+#ifndef _RDKAFKACPP_H_
+#define _RDKAFKACPP_H_
 
 /**
  * @file rdkafkacpp.h
@@ -69,6 +70,13 @@
 
 /**@endcond*/
 
+extern "C" {
+        /* Forward declarations */
+        struct rd_kafka_s;
+        struct rd_kafka_topic_s;
+        struct rd_kafka_message_s;
+};
+
 namespace RdKafka {
 
 
@@ -91,7 +99,7 @@ namespace RdKafka {
  * @remark This value should only be used during compile time,
  *         for runtime checks of version use RdKafka::version()
  */
-#define RD_KAFKA_VERSION  0x000b00ff
+#define RD_KAFKA_VERSION  0x000b05ff
 
 /**
  * @brief Returns the librdkafka version as integer.
@@ -236,7 +244,18 @@ enum ErrorCode {
         ERR__KEY_DESERIALIZATION = -160,
         /** Value deserialization error */
         ERR__VALUE_DESERIALIZATION = -159,
-	/** End internal error codes */
+        /** Partial response */
+        ERR__PARTIAL = -158,
+        /** Modification attempted on read-only object */
+        ERR__READ_ONLY = -157,
+        /** No such entry / item not found */
+        ERR__NOENT = -156,
+        /** Read underflow */
+        ERR__UNDERFLOW = -155,
+        /** Invalid type */
+        ERR__INVALID_TYPE = -154,
+
+        /** End internal error codes */
 	ERR__END = -100,
 
 	/* Kafka broker errors: */
@@ -1141,6 +1160,41 @@ class RD_EXPORT Handle {
    *          retrieved in the allotted timespan.
    */
   virtual const std::string clusterid (int timeout_ms) = 0;
+
+  /**
+   * @brief Returns the underlying librdkafka C rd_kafka_t handle.
+   *
+   * @warning Calling the C API on this handle is not recommended and there
+   *          is no official support for it, but for cases where the C++
+   *          does not provide the proper functionality this C handle can be
+   *          used to interact directly with the core librdkafka API.
+   *
+   * @remark The lifetime of the returned pointer is the same as the Topic
+   *         object this method is called on.
+   *
+   * @remark Include <rdkafka/rdkafka.h> prior to including
+   *         <rdkafka/rdkafkacpp.h>
+   *
+   * @returns \c rd_kafka_t*
+   */
+  virtual struct rd_kafka_s *c_ptr () = 0;
+
+  /**
+   * @brief Returns the current ControllerId (controller broker id)
+   *        as reported in broker metadata.
+   *
+   * @param timeout_ms If there is no cached value from metadata retrieval
+   *                   then this specifies the maximum amount of time
+   *                   (in milliseconds) the call will block waiting
+   *                   for metadata to be retrieved.
+   *                   Use 0 for non-blocking calls.
+   *
+   * @remark Requires broker version >=0.10.0 and api.version.request=true.
+   *
+   * @returns Last cached ControllerId, or -1 if no ControllerId could be
+   *          retrieved in the allotted timespan.
+   */
+  virtual int32_t controllerid (int timeout_ms) = 0;
 };
 
 
@@ -1257,6 +1311,24 @@ class RD_EXPORT Topic {
    *          offsets could be stored.
    */
   virtual ErrorCode offset_store (int32_t partition, int64_t offset) = 0;
+
+  /**
+   * @brief Returns the underlying librdkafka C rd_kafka_topic_t handle.
+   *
+   * @warning Calling the C API on this handle is not recommended and there
+   *          is no official support for it, but for cases where the C++ API
+   *          does not provide the underlying functionality this C handle can be
+   *          used to interact directly with the core librdkafka API.
+   *
+   * @remark The lifetime of the returned pointer is the same as the Topic
+   *         object this method is called on.
+   *
+   * @remark Include <rdkafka/rdkafka.h> prior to including
+   *         <rdkafka/rdkafkacpp.h>
+   *
+   * @returns \c rd_kafka_topic_t*
+   */
+  virtual struct rd_kafka_topic_s *c_ptr () = 0;
 };
 
 
@@ -1361,6 +1433,24 @@ class RD_EXPORT Message {
   /** @returns the latency in microseconds for a produced message measured
    *           from the produce() call, or -1 if latency is not available. */
   virtual int64_t             latency () const = 0;
+
+  /**
+   * @brief Returns the underlying librdkafka C rd_kafka_message_t handle.
+   *
+   * @warning Calling the C API on this handle is not recommended and there
+   *          is no official support for it, but for cases where the C++ API
+   *          does not provide the underlying functionality this C handle can be
+   *          used to interact directly with the core librdkafka API.
+   *
+   * @remark The lifetime of the returned pointer is the same as the Message
+   *         object this method is called on.
+   *
+   * @remark Include <rdkafka/rdkafka.h> prior to including
+   *         <rdkafka/rdkafkacpp.h>
+   *
+   * @returns \c rd_kafka_message_t*
+   */
+  virtual struct rd_kafka_message_s *c_ptr () = 0;
 };
 
 /**@}*/
@@ -1429,6 +1519,23 @@ class RD_EXPORT Queue {
   virtual int poll (int timeout_ms) = 0;
 
   virtual ~Queue () = 0;
+
+  /**
+   * @brief Enable IO event triggering for queue.
+   *
+   * To ease integration with IO based polling loops this API
+   * allows an application to create a separate file-descriptor
+   * that librdkafka will write \p payload (of size \p size) to
+   * whenever a new element is enqueued on a previously empty queue.
+   *
+   * To remove event triggering call with \p fd = -1.
+   *
+   * librdkafka will maintain a copy of the \p payload.
+   *
+   * @remark When using forwarded queues the IO event must only be enabled
+   *         on the final forwarded-to (destination) queue.
+   */
+  virtual void io_event_enable (int fd, const void *payload, size_t size) = 0;
 };
 
 /**@}*/
@@ -1701,7 +1808,10 @@ public:
    *
    * @remark \c enable.auto.offset.store must be set to \c false when using this API.
    *
-   * @returns RdKafka::ERR_NO_ERROR on success or an error code on error.
+   * @returns RdKafka::ERR_NO_ERROR on success, or
+   *          RdKafka::ERR___UNKNOWN_PARTITION if none of the offsets could
+   *          be stored, or
+   *          RdKafka::ERR___INVALID_ARG if \c enable.auto.offset.store is true.
    */
   virtual ErrorCode offsets_store (std::vector<TopicPartition*> &offsets) = 0;
 };
@@ -1983,7 +2093,7 @@ class RD_EXPORT Producer : public virtual Handle {
    *
    *  NOTE: RK_MSG_FREE and RK_MSG_COPY are mutually exclusive.
    *
-   *  If the function returns -1 and RK_MSG_FREE was specified, then
+   *  If the function returns an error code and RK_MSG_FREE was specified, then
    *  the memory associated with the payload is still the caller's
    *  responsibility.
    *
@@ -1998,8 +2108,10 @@ class RD_EXPORT Producer : public virtual Handle {
    * referencing this message.
    *
    * @returns an ErrorCode to indicate success or failure:
-   *  - ERR__QUEUE_FULL - maximum number of outstanding messages has been
-   *                      reached: \c queue.buffering.max.message
+   *  - ERR_NO_ERROR           - message successfully enqueued for transmission.
+   *
+   *  - ERR__QUEUE_FULL        - maximum number of outstanding messages has been
+   *                             reached: \c queue.buffering.max.message
    *
    *  - ERR_MSG_SIZE_TOO_LARGE - message is larger than configured max size:
    *                            \c messages.max.bytes
@@ -2188,3 +2300,4 @@ class Metadata {
 
 }
 
+#endif /* _RDKAFKACPP_H_ */
