@@ -40,6 +40,10 @@
 #include "rdkafka_plugin.h"
 #endif
 
+#ifndef _MSC_VER
+#include <netinet/tcp.h>
+#endif
+
 struct rd_kafka_property {
 	rd_kafka_conf_scope_t scope;
 	const char *name;
@@ -120,6 +124,19 @@ rd_kafka_conf_validate_single (const struct rd_kafka_property *prop,
 	return !strchr(val, ',') && !strchr(val, ' ');
 }
 
+/**
+ * @brief Validate builtin partitioner string
+ */
+static RD_UNUSED int
+rd_kafka_conf_validate_partitioner (const struct rd_kafka_property *prop,
+                                    const char *val, int ival) {
+        return !strcmp(val, "random") ||
+                !strcmp(val, "consistent") ||
+                !strcmp(val, "consistent_random") ||
+                !strcmp(val, "murmur2") ||
+                !strcmp(val, "murmur2_random");
+}
+
 
 /**
  * librdkafka configuration property definitions.
@@ -178,12 +195,12 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
 	  0, 1000000000, 0xffff },
 	{ _RK_GLOBAL, "receive.message.max.bytes", _RK_C_INT,
           _RK(recv_max_msg_size),
-	  "Maximum Kafka protocol response message size. "
-	  "This is a safety precaution to avoid memory exhaustion in case of "
-	  "protocol hickups. "
-          "The value should be at least fetch.message.max.bytes * number of "
-          "partitions consumed from + messaging overhead (e.g. 200000 bytes).",
-	  1000, 1000000000, 100000000 },
+          "Maximum Kafka protocol response message size. "
+          "This serves as a safety precaution to avoid memory exhaustion in "
+          "case of protocol hickups. "
+          "This value is automatically adjusted upwards to be at least "
+          "`fetch.max.bytes` + 512 to allow for protocol overhead.",
+	  1000, INT_MAX, 100000000 },
 	{ _RK_GLOBAL, "max.in.flight.requests.per.connection", _RK_C_INT,
 	  _RK(max_inflight),
 	  "Maximum number of in-flight requests per broker connection. "
@@ -208,7 +225,7 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
 	{ _RK_GLOBAL, "metadata.max.age.ms", _RK_C_INT,
           _RK(metadata_max_age_ms),
           "Metadata cache max age. "
-          "Defaults to metadata.refresh.interval.ms * 3",
+          "Defaults to topic.metadata.refresh.interval.ms * 3",
           1, 24*3600*1000, -1 },
         { _RK_GLOBAL, "topic.metadata.refresh.fast.interval.ms", _RK_C_INT,
           _RK(metadata_refresh_fast_interval_ms),
@@ -232,25 +249,35 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
           "broker metadata information as if the topics did not exist." },
 	{ _RK_GLOBAL, "debug", _RK_C_S2F, _RK(debug),
 	  "A comma-separated list of debug contexts to enable. "
-	  "Debugging the Producer: broker,topic,msg. Consumer: cgrp,topic,fetch",
+          "Detailed Producer debugging: broker,topic,msg. "
+          "Consumer: consumer,cgrp,topic,fetch",
 	  .s2i = {
                         { RD_KAFKA_DBG_GENERIC,  "generic" },
 			{ RD_KAFKA_DBG_BROKER,   "broker" },
 			{ RD_KAFKA_DBG_TOPIC,    "topic" },
 			{ RD_KAFKA_DBG_METADATA, "metadata" },
+                        { RD_KAFKA_DBG_FEATURE,  "feature" },
 			{ RD_KAFKA_DBG_QUEUE,    "queue" },
 			{ RD_KAFKA_DBG_MSG,      "msg" },
 			{ RD_KAFKA_DBG_PROTOCOL, "protocol" },
                         { RD_KAFKA_DBG_CGRP,     "cgrp" },
 			{ RD_KAFKA_DBG_SECURITY, "security" },
 			{ RD_KAFKA_DBG_FETCH,    "fetch" },
-			{ RD_KAFKA_DBG_FEATURE,  "feature" },
                         { RD_KAFKA_DBG_INTERCEPTOR, "interceptor" },
                         { RD_KAFKA_DBG_PLUGIN,   "plugin" },
-			{ RD_KAFKA_DBG_ALL,      "all" },
+                        { RD_KAFKA_DBG_CONSUMER, "consumer" },
+                        { RD_KAFKA_DBG_ADMIN,    "admin" },
+			{ RD_KAFKA_DBG_ALL,      "all" }
 		} },
 	{ _RK_GLOBAL, "socket.timeout.ms", _RK_C_INT, _RK(socket_timeout_ms),
-	  "Timeout for network requests.",
+	  "Default timeout for network requests. "
+          "Producer: ProduceRequests will use the lesser value of "
+          "`socket.timeout.ms` and remaining `message.timeout.ms` for the "
+          "first message in the batch. "
+          "Consumer: FetchRequests will use "
+          "`fetch.wait.max.ms` + `socket.timeout.ms`. "
+          "Admin: Admin requests will use `socket.timeout.ms` or explicitly "
+          "set `rd_kafka_AdminOptions_set_operation_timeout()` value.",
 	  10, 300*1000, 60*1000 },
 	{ _RK_GLOBAL, "socket.blocking.max.ms", _RK_C_INT,
 	  _RK(socket_blocking_max_ms),
@@ -266,20 +293,24 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
 	  _RK(socket_rcvbuf_size),
 	  "Broker socket receive buffer size. System default is used if 0.",
 	  0, 100000000, 0 },
+#ifdef SO_KEEPALIVE
 	{ _RK_GLOBAL, "socket.keepalive.enable", _RK_C_BOOL,
 	  _RK(socket_keepalive),
           "Enable TCP keep-alives (SO_KEEPALIVE) on broker sockets",
           0, 1, 0 },
+#endif
+#ifdef TCP_NODELAY
 	{ _RK_GLOBAL, "socket.nagle.disable", _RK_C_BOOL,
 	  _RK(socket_nagle_disable),
-          "Disable the Nagle algorithm (TCP_NODELAY).",
+          "Disable the Nagle algorithm (TCP_NODELAY) on broker sockets.",
           0, 1, 0 },
+#endif
         { _RK_GLOBAL, "socket.max.fails", _RK_C_INT,
           _RK(socket_max_fails),
           "Disconnect from broker when this number of send failures "
           "(e.g., timed out requests) is reached. Disable with 0. "
           "NOTE: The connection is automatically re-established.",
-          0, 1000000, 3 },
+          0, 1000000, 1 },
 	{ _RK_GLOBAL, "broker.address.ttl", _RK_C_INT,
 	  _RK(broker_addr_ttl),
 	  "How long to cache the broker address resolving "
@@ -344,6 +375,10 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
           "It might be useful to turn this off when interacting with "
           "0.9 brokers with an aggressive `connection.max.idle.ms` value.",
 	  0, 1, 1 },
+        { _RK_GLOBAL, "background_event_cb", _RK_C_PTR,
+          _RK(background_event_cb),
+          "Background queue event callback "
+          "(set with rd_kafka_conf_set_background_event_cb())" },
         { _RK_GLOBAL, "socket_cb", _RK_C_PTR,
           _RK(socket_cb),
           "Socket creation callback to provide race-free CLOEXEC",
@@ -450,6 +485,22 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
 	  "protocol. See manual page for `ciphers(1)` and "
 	  "`SSL_CTX_set_cipher_list(3)."
 	},
+#if OPENSSL_VERSION_NUMBER >= 0x1000200fL
+        { _RK_GLOBAL, "ssl.curves.list", _RK_C_STR,
+          _RK(ssl.curves_list),
+          "The supported-curves extension in the TLS ClientHello message specifies "
+          "the curves (standard/named, or 'explicit' GF(2^k) or GF(p)) the client "
+          "is willing to have the server use. See manual page for "
+          "`SSL_CTX_set1_curves_list(3)`. OpenSSL >= 1.0.2 required."
+        },
+        { _RK_GLOBAL, "ssl.sigalgs.list", _RK_C_STR,
+          _RK(ssl.sigalgs_list),
+          "The client uses the TLS ClientHello signature_algorithms extension "
+          "to indicate to the server which signature/hash algorithm pairs "
+          "may be used in digital signatures. See manual page for "
+          "`SSL_CTX_set1_sigalgs_list(3)`. OpenSSL >= 1.0.2 required."
+        },
+#endif
 	{ _RK_GLOBAL, "ssl.key.location", _RK_C_STR,
 	  _RK(ssl.key_location),
 	  "Path to client's private key (PEM) used for authentication."
@@ -471,16 +522,18 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
 	  _RK(ssl.crl_location),
 	  "Path to CRL for verifying broker's certificate validity."
 	},
+	{ _RK_GLOBAL, "ssl.keystore.location", _RK_C_STR,
+	_RK(ssl.keystore_location),
+	"Path to client's keystore (PKCS#12) used for authentication."
+	},
+	{ _RK_GLOBAL, "ssl.keystore.password", _RK_C_STR,
+	_RK(ssl.keystore_password),
+	"Client's keystore (PKCS#12) password."
+	},
 #endif /* WITH_SSL */
 
         /* Point user in the right direction if they try to apply
          * Java client SSL / JAAS properties. */
-        { _RK_GLOBAL, "ssl.keystore.location", _RK_C_INVALID,
-          _RK(dummy),
-          "Java KeyStores are not supported, use `ssl.key.location` and "
-          "a private key (PEM) file instead. "
-          "See https://github.com/edenhill/librdkafka/wiki/Using-SSL-with-librdkafka for more information."
-        },
         { _RK_GLOBAL, "ssl.truststore.location", _RK_C_INVALID,
           _RK(dummy),
           "Java TrustStores are not supported, use `ssl.ca.location` "
@@ -501,6 +554,8 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
 	 "**NOTE**: Despite the name only one mechanism must be configured.",
 	 .sdef = "GSSAPI",
 	 .validate = rd_kafka_conf_validate_single },
+        {_RK_GLOBAL,"sasl.mechanism", _RK_C_ALIAS,
+         .sdef = "sasl.mechanisms" },
 	{ _RK_GLOBAL, "sasl.kerberos.service.name", _RK_C_STR,
 	  _RK(sasl.service_name),
 	  "Kerberos principal name that Kafka runs as, "
@@ -593,7 +648,11 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
         /* Global consumer properties */
         { _RK_GLOBAL|_RK_CONSUMER, "enable.auto.commit", _RK_C_BOOL,
           _RK(enable_auto_commit),
-          "Automatically and periodically commit offsets in the background.",
+          "Automatically and periodically commit offsets in the background. "
+          "Note: setting this to false does not prevent the consumer from "
+          "fetching previously committed start offsets. To circumvent this "
+          "behaviour set specific start offsets per partition in the call "
+          "to assign().",
           0, 1, 1 },
         { _RK_GLOBAL|_RK_CONSUMER, "auto.commit.interval.ms", _RK_C_INT,
 	  _RK(auto_commit_interval_ms),
@@ -633,6 +692,19 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
           1, 1000000000, 1024*1024 },
 	{ _RK_GLOBAL|_RK_CONSUMER, "max.partition.fetch.bytes", _RK_C_ALIAS,
 	  .sdef = "fetch.message.max.bytes" },
+        { _RK_GLOBAL|_RK_CONSUMER, "fetch.max.bytes", _RK_C_INT,
+          _RK(fetch_max_bytes),
+          "Maximum amount of data the broker shall return for a Fetch request. "
+          "Messages are fetched in batches by the consumer and if the first "
+          "message batch in the first non-empty partition of the Fetch request "
+          "is larger than this value, then the message batch will still be "
+          "returned to ensure the consumer can make progress. "
+          "The maximum message batch size accepted by the broker is defined "
+          "via `message.max.bytes` (broker config) or "
+          "`max.message.bytes` (broker topic config). "
+          "`fetch.max.bytes` is automatically adjusted upwards to be "
+          "at least `message.max.bytes` (consumer config).",
+          0, INT_MAX-512, 50*1024*1024 /* 50MB */ },
 	{ _RK_GLOBAL|_RK_CONSUMER, "fetch.min.bytes", _RK_C_INT,
 	  _RK(fetch_min_bytes),
 	  "Minimum number of bytes the broker responds with. "
@@ -709,8 +781,21 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
                 .sdef = "message.send.max.retries" },
 	{ _RK_GLOBAL|_RK_PRODUCER, "retry.backoff.ms", _RK_C_INT,
 	  _RK(retry_backoff_ms),
-	  "The backoff time in milliseconds before retrying a message send.",
+	  "The backoff time in milliseconds before retrying a protocol request.",
 	  1, 300*1000, 100 },
+
+	{ _RK_GLOBAL|_RK_PRODUCER, "queue.buffering.backpressure.threshold",
+          _RK_C_INT, _RK(queue_backpressure_thres),
+          "The threshold of outstanding not yet transmitted broker requests "
+          "needed to backpressure the producer's message accumulator. "
+          "If the number of not yet transmitted requests equals or exceeds "
+          "this number, produce request creation that would have otherwise "
+          "been triggered (for example, in accordance with linger.ms) will be "
+          "delayed. A lower number yields larger and more effective batches. "
+          "A higher value can improve latency when using compression on slow "
+          "machines.",
+        1, 1000000, 1 },
+
 	{ _RK_GLOBAL|_RK_PRODUCER, "compression.codec", _RK_C_S2I,
 	  _RK(compression_codec),
 	  "compression codec to use for compressing message sets. "
@@ -728,6 +813,8 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
                         { RD_KAFKA_COMPRESSION_LZ4, "lz4" },
 			{ 0 }
 		} },
+        { _RK_GLOBAL|_RK_PRODUCER, "compression.type", _RK_C_ALIAS,
+          .sdef = "compression.codec" },
 	{ _RK_GLOBAL|_RK_PRODUCER, "batch.num.messages", _RK_C_INT,
 	  _RK(batch_num_messages),
 	  "Maximum number of messages batched in one MessageSet. "
@@ -757,7 +844,7 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
 	  "*0*=Broker does not send any response/ack to client, "
 	  "*1*=Only the leader broker will need to ack the message, "
 	  "*-1* or *all*=broker will block until message is committed by all "
-	  "in sync replicas (ISRs) or broker's `in.sync.replicas` "
+	  "in sync replicas (ISRs) or broker's `min.insync.replicas` "
 	  "setting before sending response. ",
 	  -1, 1000, 1,
 	  .s2i = {
@@ -778,24 +865,54 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
 	  "Local message timeout. "
 	  "This value is only enforced locally and limits the time a "
 	  "produced message waits for successful delivery. "
-          "A time of 0 is infinite.",
+	  "A time of 0 is infinite. "
+	  "This is the maximum time librdkafka may use to deliver a message "
+	  "(including retries). Delivery error occurs when either the retry "
+	  "count or the message timeout are exceeded.",
 	  0, 900*1000, 300*1000 },
+        { _RK_TOPIC|_RK_PRODUCER, "queuing.strategy", _RK_C_S2I,
+          _RKT(queuing_strategy),
+          "Producer queuing strategy. FIFO preserves produce ordering, "
+          "while LIFO prioritizes new messages. "
+          "WARNING: `lifo` is experimental and subject to change or removal.",
+          .vdef = 0,
+          .s2i = {
+                        { RD_KAFKA_QUEUE_FIFO, "fifo" },
+                        { RD_KAFKA_QUEUE_LIFO, "lifo" }
+                }
+        },
         { _RK_TOPIC|_RK_PRODUCER, "produce.offset.report", _RK_C_BOOL,
           _RKT(produce_offset_report),
           "Report offset of produced message back to application. "
           "The application must be use the `dr_msg_cb` to retrieve the offset "
           "from `rd_kafka_message_t.offset`.",
           0, 1, 0 },
+        { _RK_TOPIC|_RK_PRODUCER, "partitioner", _RK_C_STR,
+          _RKT(partitioner_str),
+          "Partitioner: "
+          "`random` - random distribution, "
+          "`consistent` - CRC32 hash of key (Empty and NULL keys are mapped to single partition), "
+          "`consistent_random` - CRC32 hash of key (Empty and NULL keys are randomly partitioned), "
+          "`murmur2` - Java Producer compatible Murmur2 hash of key (NULL keys are mapped to single partition), "
+          "`murmur2_random` - Java Producer compatible Murmur2 hash of key (NULL keys are randomly partitioned. This is functionally equivalent to the default partitioner in the Java Producer.).",
+          .sdef = "consistent_random",
+          .validate = rd_kafka_conf_validate_partitioner },
 	{ _RK_TOPIC|_RK_PRODUCER, "partitioner_cb", _RK_C_PTR,
 	  _RKT(partitioner),
-	  "Partitioner callback "
+	  "Custom partitioner callback "
 	  "(set with rd_kafka_topic_conf_set_partitioner_cb())" },
+	{ _RK_TOPIC|_RK_PRODUCER, "msg_order_cmp", _RK_C_PTR,
+	  _RKT(msg_order_cmp),
+	  "Message queue ordering comparator "
+	  "(set with rd_kafka_topic_conf_set_msg_order_cmp()). "
+          "Also see `queuing.strategy`." },
 	{ _RK_TOPIC, "opaque", _RK_C_PTR,
 	  _RKT(opaque),
 	  "Application opaque (set with rd_kafka_topic_conf_set_opaque())" },
 	{ _RK_TOPIC | _RK_PRODUCER, "compression.codec", _RK_C_S2I,
 	  _RKT(compression_codec),
-	  "Compression codec to use for compressing message sets. ",
+	  "Compression codec to use for compressing message sets. "
+          "inherit = inherit global compression.codec configuration.",
 	  .vdef = RD_KAFKA_COMPRESSION_INHERIT,
 	  .s2i = {
 		  { RD_KAFKA_COMPRESSION_NONE, "none" },
@@ -809,6 +926,18 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
 		  { RD_KAFKA_COMPRESSION_INHERIT, "inherit" },
 		  { 0 }
 		} },
+        { _RK_TOPIC | _RK_PRODUCER, "compression.type", _RK_C_ALIAS,
+          .sdef = "compression.codec" },
+	{ _RK_TOPIC|_RK_PRODUCER, "compression.level", _RK_C_INT,
+	  _RKT(compression_level),
+	  "Compression level parameter for algorithm selected by configuration "
+	  "property `compression.codec`. Higher values will result in better "
+	  "compression at the cost of more CPU usage. Usable range is "
+	  "algorithm-dependent: [0-9] for gzip; [0-12] for lz4; only 0 for snappy; "
+	  "-1 = codec-dependent default compression level.",
+	  RD_KAFKA_COMPLEVEL_MIN,
+	  RD_KAFKA_COMPLEVEL_MAX,
+	  RD_KAFKA_COMPLEVEL_DEFAULT },
 
 
         /* Topic consumer properties */
@@ -1250,8 +1379,9 @@ rd_kafka_anyconf_set_prop (int scope, void *conf,
 
 			/* No match */
 			rd_snprintf(errstr, errstr_size,
-				 "Invalid value for "
-				 "configuration property \"%s\"", prop->name);
+                                "Invalid value \"%.*s\" for "
+                                "configuration property \"%s\"",
+                                (int)(t-s), s, prop->name);
 			return RD_KAFKA_CONF_INVALID;
 
 		}
@@ -1621,11 +1751,25 @@ rd_kafka_topic_conf_t *rd_kafka_topic_conf_dup (const rd_kafka_topic_conf_t
 	return new;
 }
 
+rd_kafka_topic_conf_t *rd_kafka_default_topic_conf_dup (rd_kafka_t *rk) {
+        if (rk->rk_conf.topic_conf)
+                return rd_kafka_topic_conf_dup(rk->rk_conf.topic_conf);
+        else
+                return rd_kafka_topic_conf_new();
+}
 
 void rd_kafka_conf_set_events (rd_kafka_conf_t *conf, int events) {
 	conf->enabled_events = events;
 }
 
+void
+rd_kafka_conf_set_background_event_cb (rd_kafka_conf_t *conf,
+                                       void (*event_cb) (rd_kafka_t *rk,
+                                                         rd_kafka_event_t *rkev,
+                                                         void *opaque)) {
+        conf->background_event_cb = event_cb;
+}
+
 
 void rd_kafka_conf_set_dr_cb (rd_kafka_conf_t *conf,
 			      void (*dr_cb) (rd_kafka_t *rk,
@@ -1767,6 +1911,15 @@ rd_kafka_topic_conf_set_partitioner_cb (rd_kafka_topic_conf_t *topic_conf,
 	topic_conf->partitioner = partitioner;
 }
 
+void
+rd_kafka_topic_conf_set_msg_order_cmp (rd_kafka_topic_conf_t *topic_conf,
+                                       int (*msg_order_cmp) (
+                                               const rd_kafka_message_t *a,
+                                               const rd_kafka_message_t *b)) {
+        topic_conf->msg_order_cmp =
+                (int (*)(const void *, const void *))msg_order_cmp;
+}
+
 void rd_kafka_topic_conf_set_opaque (rd_kafka_topic_conf_t *topic_conf,
 				     void *opaque) {
 	topic_conf->opaque = opaque;
@@ -2151,3 +2304,206 @@ void rd_kafka_conf_properties_show (FILE *fp) {
 	fprintf(fp, "\n");
         fprintf(fp, "### C/P legend: C = Consumer, P = Producer, * = both\n");
 }
+
+
+
+
+/**
+ * @name Configuration value methods
+ *
+ * @remark This generic interface will eventually replace the config property
+ *         used above.
+ * @{
+ */
+
+
+/**
+ * @brief Set up an INT confval.
+ *
+ * @oaram name Property name, must be a const static string (will not be copied)
+ */
+void rd_kafka_confval_init_int (rd_kafka_confval_t *confval,
+                                const char *name,
+                                int vmin, int vmax, int vdef) {
+        confval->name = name;
+        confval->is_enabled = 1;
+        confval->valuetype = RD_KAFKA_CONFVAL_INT;
+        confval->u.INT.vmin = vmin;
+        confval->u.INT.vmax = vmax;
+        confval->u.INT.vdef = vdef;
+        confval->u.INT.v    = vdef;
+}
+
+/**
+ * @brief Set up a PTR confval.
+ *
+ * @oaram name Property name, must be a const static string (will not be copied)
+ */
+void rd_kafka_confval_init_ptr (rd_kafka_confval_t *confval,
+                                const char *name) {
+        confval->name = name;
+        confval->is_enabled = 1;
+        confval->valuetype = RD_KAFKA_CONFVAL_PTR;
+        confval->u.PTR = NULL;
+}
+
+/**
+ * @brief Set up but disable an intval, attempt to set this confval will fail.
+ *
+ * @oaram name Property name, must be a const static string (will not be copied)
+ */
+void rd_kafka_confval_disable (rd_kafka_confval_t *confval, const char *name) {
+        confval->name = name;
+        confval->is_enabled = 0;
+}
+
+/**
+ * @brief Set confval's value to \p valuep, verifying the passed
+ *        \p valuetype matches (or can be cast to) \p confval's type.
+ *
+ * @param dispname is the display name for the configuration value and is
+ *        included in error strings.
+ * @param valuep is a pointer to the value, or NULL to revert to default.
+ *
+ * @returns RD_KAFKA_RESP_ERR_NO_ERROR if the new value was set, or
+ *          RD_KAFKA_RESP_ERR__INVALID_ARG if the value was of incorrect type,
+ *          out of range, or otherwise not a valid value.
+ */
+rd_kafka_resp_err_t
+rd_kafka_confval_set_type (rd_kafka_confval_t *confval,
+                           rd_kafka_confval_type_t valuetype,
+                           const void *valuep,
+                           char *errstr, size_t errstr_size) {
+
+        if (!confval->is_enabled) {
+                rd_snprintf(errstr, errstr_size,
+                            "\"%s\" is not supported for this operation",
+                            confval->name);
+                return RD_KAFKA_RESP_ERR__INVALID_ARG;
+        }
+
+        switch (confval->valuetype)
+        {
+        case RD_KAFKA_CONFVAL_INT:
+        {
+                int v;
+                const char *end;
+
+                if (!valuep) {
+                        /* Revert to default */
+                        confval->u.INT.v = confval->u.INT.vdef;
+                        confval->is_set = 0;
+                        return RD_KAFKA_RESP_ERR_NO_ERROR;
+                }
+
+                switch (valuetype)
+                {
+                case RD_KAFKA_CONFVAL_INT:
+                        v = *(const int *)valuep;
+                        break;
+                case RD_KAFKA_CONFVAL_STR:
+                        v = (int)strtol((const char *)valuep, (char **)&end, 0);
+                        if (end == (const char *)valuep) {
+                                rd_snprintf(errstr, errstr_size,
+                                            "Invalid value type for \"%s\": "
+                                            "expecting integer",
+                                            confval->name);
+                                return RD_KAFKA_RESP_ERR__INVALID_TYPE;
+                        }
+                default:
+                        rd_snprintf(errstr, errstr_size,
+                                    "Invalid value type for \"%s\": "
+                                    "expecting integer", confval->name);
+                        return RD_KAFKA_RESP_ERR__INVALID_ARG;
+                }
+
+
+                if ((confval->u.INT.vmin || confval->u.INT.vmax) &&
+                    (v < confval->u.INT.vmin || v > confval->u.INT.vmax)) {
+                        rd_snprintf(errstr, errstr_size,
+                                    "Invalid value type for \"%s\": "
+                                    "expecting integer in range %d..%d",
+                                    confval->name,
+                                    confval->u.INT.vmin,
+                                    confval->u.INT.vmax);
+                        return RD_KAFKA_RESP_ERR__INVALID_ARG;
+                }
+
+                confval->u.INT.v = v;
+                confval->is_set = 1;
+        }
+        break;
+
+        case RD_KAFKA_CONFVAL_STR:
+        {
+                size_t vlen;
+                const char *v = (const char *)valuep;
+
+                if (!valuep) {
+                        confval->is_set = 0;
+                        if (confval->u.STR.vdef)
+                                confval->u.STR.v = rd_strdup(confval->u.STR.
+                                                             vdef);
+                        else
+                                confval->u.STR.v = NULL;
+                }
+
+                if (valuetype != RD_KAFKA_CONFVAL_STR) {
+                        rd_snprintf(errstr, errstr_size,
+                                    "Invalid value type for \"%s\": "
+                                    "expecting string", confval->name);
+                        return RD_KAFKA_RESP_ERR__INVALID_ARG;
+                }
+
+                vlen = strlen(v);
+                if ((confval->u.STR.minlen || confval->u.STR.maxlen) &&
+                    (vlen < confval->u.STR.minlen ||
+                     vlen > confval->u.STR.maxlen)) {
+                        rd_snprintf(errstr, errstr_size,
+                                    "Invalid value for \"%s\": "
+                                    "expecting string with length "
+                                    "%"PRIusz"..%"PRIusz,
+                                    confval->name,
+                                    confval->u.STR.minlen,
+                                    confval->u.STR.maxlen);
+                        return RD_KAFKA_RESP_ERR__INVALID_ARG;
+                }
+
+                if (confval->u.STR.v)
+                        rd_free(confval->u.STR.v);
+
+                confval->u.STR.v = rd_strdup(v);
+        }
+        break;
+
+        case RD_KAFKA_CONFVAL_PTR:
+                confval->u.PTR = (void *)valuep;
+                break;
+
+        default:
+                RD_NOTREACHED();
+                return RD_KAFKA_RESP_ERR__NOENT;
+        }
+
+        return RD_KAFKA_RESP_ERR_NO_ERROR;
+}
+
+
+int rd_kafka_confval_get_int (const rd_kafka_confval_t *confval) {
+        rd_assert(confval->valuetype == RD_KAFKA_CONFVAL_INT);
+        return confval->u.INT.v;
+}
+
+
+const char *rd_kafka_confval_get_str (const rd_kafka_confval_t *confval) {
+        rd_assert(confval->valuetype == RD_KAFKA_CONFVAL_STR);
+        return confval->u.STR.v;
+}
+
+void *rd_kafka_confval_get_ptr (const rd_kafka_confval_t *confval) {
+        rd_assert(confval->valuetype == RD_KAFKA_CONFVAL_PTR);
+        return confval->u.PTR;
+}
+
+
+/**@}*/
