@@ -59,6 +59,7 @@
 #include "rdkafka_msgset.h"
 #include "rdkafka_topic.h"
 #include "rdkafka_partition.h"
+#include "rdkafka_header.h"
 #include "rdkafka_lz4.h"
 
 #include "rdvarint.h"
@@ -116,12 +117,29 @@ typedef struct rd_kafka_msgset_reader_s {
                                           *          reference! */
 
         int          msetr_msgcnt;      /**< Number of messages in rkq */
+        int64_t      msetr_msg_bytes;   /**< Number of bytes in rkq */
         rd_kafka_q_t msetr_rkq;         /**< Temp Message and error queue */
         rd_kafka_q_t *msetr_par_rkq;    /**< Parent message and error queue,
                                          *   the temp msetr_rkq will be moved
                                          *   to this queue when parsing
                                          *   is done.
                                          *   Refcount is not increased. */
+
+        int64_t msetr_next_offset;      /**< Next offset to fetch after
+                                         *   this reader run is done.
+                                         *   Optional: only used for special
+                                         *   cases where the per-message offset
+                                         *   can't be relied on for next
+                                         *   fetch offset, such as with
+                                         *   compacted topics. */
+
+        const char *msetr_srcname;      /**< Optional message source string,
+                                         *   used in debug logging to
+                                         *   indicate messages were
+                                         *   from an inner compressed
+                                         *   message set.
+                                         *   Not freed (use const memory).
+                                         *   Add trailing space. */
 } rd_kafka_msgset_reader_t;
 
 
@@ -149,6 +167,9 @@ rd_kafka_msgset_reader_init (rd_kafka_msgset_reader_t *msetr,
         msetr->msetr_rktp       = rktp;
         msetr->msetr_tver       = tver;
         msetr->msetr_rkbuf      = rkbuf;
+        msetr->msetr_srcname    = "";
+
+        rkbuf->rkbuf_uflow_mitigation = "truncated response from broker (ok)";
 
         /* All parsed messages are put on this temporary op
          * queue first and then moved in one go to the real op queue. */
@@ -366,6 +387,8 @@ rd_kafka_msgset_reader_decompress (rd_kafka_msgset_reader_t *msetr,
                                             msetr->msetr_tver,
                                             &msetr->msetr_rkq);
 
+                inner_msetr.msetr_srcname = "compressed ";
+
                 if (MsgVersion == 1) {
                         /* postproc() will convert relative to
                          * absolute offsets */
@@ -384,11 +407,18 @@ rd_kafka_msgset_reader_decompress (rd_kafka_msgset_reader_t *msetr,
                 /* Parse the inner MessageSet */
                 err = rd_kafka_msgset_reader_run(&inner_msetr);
 
+                /* Transfer message count from inner to outer */
+                msetr->msetr_msgcnt += inner_msetr.msetr_msgcnt;
+                msetr->msetr_msg_bytes += inner_msetr.msetr_msg_bytes;
+
 
         } else {
                 /* MsgVersion 2 */
                 rd_kafka_buf_t *orig_rkbuf = msetr->msetr_rkbuf;
 
+                rkbufz->rkbuf_uflow_mitigation =
+                        "truncated response from broker (ok)";
+
                 /* Temporarily replace read buffer with uncompressed buffer */
                 msetr->msetr_rkbuf = rkbufz;
 
@@ -587,6 +617,7 @@ rd_kafka_msgset_reader_msg_v0_1 (rd_kafka_msgset_reader_t *msetr) {
         /* Enqueue message on temporary queue */
         rd_kafka_q_enq(&msetr->msetr_rkq, rko);
         msetr->msetr_msgcnt++;
+        msetr->msetr_msg_bytes += rkm->rkm_key_len + rkm->rkm_len;
 
         return RD_KAFKA_RESP_ERR_NO_ERROR; /* Continue */
 
@@ -608,13 +639,13 @@ rd_kafka_msgset_reader_msg_v2 (rd_kafka_msgset_reader_t *msetr) {
         rd_kafka_toppar_t *rktp = msetr->msetr_rktp;
         struct {
                 int64_t Length;
-                int64_t  MsgAttributes; /* int8_t, but int64 req. for varint */
+                int8_t  MsgAttributes;
                 int64_t TimestampDelta;
                 int64_t OffsetDelta;
                 int64_t Offset;  /* Absolute offset */
                 rd_kafkap_bytes_t Key;
                 rd_kafkap_bytes_t Value;
-                int64_t HeaderCnt;
+                rd_kafkap_bytes_t Headers;
         } hdr;
         rd_kafka_op_t *rko;
         rd_kafka_msg_t *rkm;
@@ -625,7 +656,7 @@ rd_kafka_msgset_reader_msg_v2 (rd_kafka_msgset_reader_t *msetr) {
 
         rd_kafka_buf_read_varint(rkbuf, &hdr.Length);
         message_end = rd_slice_offset(&rkbuf->rkbuf_reader)+(size_t)hdr.Length;
-        rd_kafka_buf_read_varint(rkbuf, &hdr.MsgAttributes);
+        rd_kafka_buf_read_i8(rkbuf, &hdr.MsgAttributes);
 
         rd_kafka_buf_read_varint(rkbuf, &hdr.TimestampDelta);
         rd_kafka_buf_read_varint(rkbuf, &hdr.OffsetDelta);
@@ -634,7 +665,10 @@ rd_kafka_msgset_reader_msg_v2 (rd_kafka_msgset_reader_t *msetr) {
         /* Skip message if outdated */
         if (hdr.Offset < rktp->rktp_offsets.fetch_offset) {
                 rd_rkb_dbg(msetr->msetr_rkb, MSG, "MSG",
+                           "%s [%"PRId32"]: "
                            "Skip offset %"PRId64" < fetch_offset %"PRId64,
+                           rktp->rktp_rkt->rkt_topic->str,
+                           rktp->rktp_partition,
                            hdr.Offset, rktp->rktp_offsets.fetch_offset);
                 rd_kafka_buf_skip_to(rkbuf, message_end);
                 return RD_KAFKA_RESP_ERR_NO_ERROR; /* Continue with next msg */
@@ -644,8 +678,11 @@ rd_kafka_msgset_reader_msg_v2 (rd_kafka_msgset_reader_t *msetr) {
 
         rd_kafka_buf_read_bytes_varint(rkbuf, &hdr.Value);
 
-        /* Ignore headers for now */
-        rd_kafka_buf_skip_to(rkbuf, message_end);
+        /* We parse the Headers later, just store the size (possibly truncated)
+         * and pointer to the headers. */
+        hdr.Headers.len = (int32_t)(message_end -
+                                    rd_slice_offset(&rkbuf->rkbuf_reader));
+        rd_kafka_buf_read_ptr(rkbuf, &hdr.Headers.data, hdr.Headers.len);
 
         /* Create op/message container for message. */
         rko = rd_kafka_op_new_fetch_msg(&rkm,
@@ -658,6 +695,13 @@ rd_kafka_msgset_reader_msg_v2 (rd_kafka_msgset_reader_t *msetr) {
                                         RD_KAFKAP_BYTES_IS_NULL(&hdr.Value) ?
                                         NULL : hdr.Value.data);
 
+        /* Store pointer to unparsed message headers, they will
+         * be parsed on the first access.
+         * This pointer points to the rkbuf payload.
+         * Note: can't perform struct copy here due to const fields (MSVC) */
+        rkm->rkm_u.consumer.binhdrs.len  = hdr.Headers.len;
+        rkm->rkm_u.consumer.binhdrs.data = hdr.Headers.data;
+
         /* Set timestamp.
          *
          * When broker assigns the timestamps (LOG_APPEND_TIME) it will
@@ -679,6 +723,7 @@ rd_kafka_msgset_reader_msg_v2 (rd_kafka_msgset_reader_t *msetr) {
         /* Enqueue message on temporary queue */
         rd_kafka_q_enq(&msetr->msetr_rkq, rko);
         msetr->msetr_msgcnt++;
+        msetr->msetr_msg_bytes += rkm->rkm_key_len + rkm->rkm_len;
 
         return RD_KAFKA_RESP_ERR_NO_ERROR;
 
@@ -791,15 +836,13 @@ rd_kafka_msgset_reader_v2 (rd_kafka_msgset_reader_t *msetr) {
                                      len_start);
 
         if (unlikely(payload_size > rd_kafka_buf_read_remain(rkbuf)))
-                rd_kafka_buf_parse_fail(rkbuf,
-                                        "%s [%"PRId32"] "
-                                        "MessageSet at offset %"PRId64
-                                        " payload size %"PRIusz
-                                        " > %"PRIusz" remaining bytes",
-                                        rktp->rktp_rkt->rkt_topic->str,
-                                        rktp->rktp_partition,
-                                        hdr.BaseOffset, payload_size,
-                                        rd_kafka_buf_read_remain(rkbuf));
+                rd_kafka_buf_underflow_fail(rkbuf, payload_size,
+                                            "%s [%"PRId32"] "
+                                            "MessageSet at offset %"PRId64
+                                            " payload size %"PRIusz,
+                                            rktp->rktp_rkt->rkt_topic->str,
+                                            rktp->rktp_partition,
+                                            hdr.BaseOffset, payload_size);
 
         /* If entire MessageSet contains old outdated offsets, skip it. */
         if (LastOffset < rktp->rktp_offsets.fetch_offset) {
@@ -856,8 +899,7 @@ rd_kafka_msgset_reader_v2 (rd_kafka_msgset_reader_t *msetr) {
          * to avoid getting stuck on compacted MessageSets where the last
          * Message in the MessageSet has an Offset < MessageSet header's
          * last offset.  See KAFKA-5443 */
-        if (likely(LastOffset >= msetr->msetr_rktp->rktp_offsets.fetch_offset))
-                msetr->msetr_rktp->rktp_offsets.fetch_offset = LastOffset + 1;
+        msetr->msetr_next_offset = LastOffset + 1;
 
         msetr->msetr_v2_hdr = NULL;
 
@@ -1030,15 +1072,18 @@ rd_kafka_msgset_reader_run (rd_kafka_msgset_reader_t *msetr) {
                 /* Ignore parse errors if there was at least one
                  * good message since it probably indicates a
                  * partial response rather than an erroneous one. */
-                if (err == RD_KAFKA_RESP_ERR__BAD_MSG &&
+                if (err == RD_KAFKA_RESP_ERR__UNDERFLOW &&
                     msetr->msetr_msgcnt > 0)
                         err = RD_KAFKA_RESP_ERR_NO_ERROR;
         }
 
         rd_rkb_dbg(msetr->msetr_rkb, MSG | RD_KAFKA_DBG_FETCH, "CONSUME",
-                   "Enqueue %i message(s) (%d ops) on %s [%"PRId32"] "
+                   "Enqueue %i %smessage(s) (%"PRId64" bytes, %d ops) on "
+                   "%s [%"PRId32"] "
                    "fetch queue (qlen %d, v%d, last_offset %"PRId64")",
-                   msetr->msetr_msgcnt, rd_kafka_q_len(&msetr->msetr_rkq),
+                   msetr->msetr_msgcnt, msetr->msetr_srcname,
+                   msetr->msetr_msg_bytes,
+                   rd_kafka_q_len(&msetr->msetr_rkq),
                    rktp->rktp_rkt->rkt_topic->str,
                    rktp->rktp_partition, rd_kafka_q_len(&msetr->msetr_rkq),
                    msetr->msetr_tver->version, last_offset);
@@ -1048,13 +1093,15 @@ rd_kafka_msgset_reader_run (rd_kafka_msgset_reader_t *msetr) {
         if (rd_kafka_q_concat(msetr->msetr_par_rkq, &msetr->msetr_rkq) != -1) {
                 /* Update partition's fetch offset based on
                  * last message's offest. */
-                if (likely(last_offset != -1)) {
+                if (likely(last_offset != -1))
                         rktp->rktp_offsets.fetch_offset = last_offset + 1;
-                        rd_atomic64_add(&rktp->rktp_c.msgs,
-                                        msetr->msetr_msgcnt);
-                }
         }
 
+        /* Adjust next fetch offset if outlier code has indicated
+         * an even later next offset. */
+        if (msetr->msetr_next_offset > rktp->rktp_offsets.fetch_offset)
+                rktp->rktp_offsets.fetch_offset = msetr->msetr_next_offset;
+
         rd_kafka_q_destroy_owner(&msetr->msetr_rkq);
 
         /* Skip remaining part of slice so caller can continue
@@ -1079,12 +1126,24 @@ rd_kafka_msgset_parse (rd_kafka_buf_t *rkbuf,
                        rd_kafka_toppar_t *rktp,
                        const struct rd_kafka_toppar_ver *tver) {
         rd_kafka_msgset_reader_t msetr;
+        rd_kafka_resp_err_t err;
 
         rd_kafka_msgset_reader_init(&msetr, rkbuf, rktp, tver,
                                     rktp->rktp_fetchq);
 
         /* Parse and handle the message set */
-        return rd_kafka_msgset_reader_run(&msetr);
+        err = rd_kafka_msgset_reader_run(&msetr);
+
+        rd_atomic64_add(&rktp->rktp_c.rx_msgs, msetr.msetr_msgcnt);
+        rd_atomic64_add(&rktp->rktp_c.rx_msg_bytes, msetr.msetr_msg_bytes);
+
+        rd_avg_add(&rktp->rktp_rkt->rkt_avg_batchcnt,
+                   (int64_t)msetr.msetr_msgcnt);
+        rd_avg_add(&rktp->rktp_rkt->rkt_avg_batchsize,
+                   (int64_t)msetr.msetr_msg_bytes);
+
+        return err;
+
 }
 
 
