@@ -25,7 +25,8 @@
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
-#pragma once
+#ifndef _TEST_H_
+#define _TEST_H_
 
 #include "../src/rd.h"
 
@@ -102,6 +103,7 @@ struct test {
 				*   total test run status. */
 #define TEST_F_MANUAL      0x4 /**< Manual test, only started when specifically
                                 *   stated */
+#define TEST_F_SOCKEM      0x8 /**< Test requires socket emulation. */
 	int minver;          /**< Limit tests to broker version range. */
 	int maxver;
 
@@ -111,6 +113,10 @@ struct test {
 	int report_cnt;
 	int report_size;
 
+        rd_kafka_resp_err_t exp_dr_err; /* Expected error in test_dr_cb */
+        int produce_sync;    /**< test_produce_sync() call in action */
+        rd_kafka_resp_err_t produce_sync_err;  /**< DR error */
+
         /**
          * Runtime
          */
@@ -120,6 +126,8 @@ struct test {
         FILE   *stats_fp;
 	int64_t timeout;
         test_state_t state;
+        int     failcnt;     /**< Number of failures, useful with FAIL_LATER */
+        char    failstr[512];/**< First test failure reason */
 
 #if WITH_SOCKEM
         rd_list_t sockets;
@@ -142,32 +150,12 @@ struct test {
 #define TEST_F_KNOWN_ISSUE_OSX  0
 #endif
 
+void test_fail0 (const char *file, int line, const char *function,
+                 int do_lock, int fail_now, const char *fmt, ...);
 
-#define TEST_FAIL0(file,line,do_lock,fail_now,...) do {                 \
-                int is_thrd = 0;                                        \
-		TEST_SAYL(0, "TEST FAILURE\n");				\
-		fprintf(stderr, "\033[31m### Test \"%s\" failed at %s:%i:%s(): ###\n", \
-			test_curr->name,                                \
-                        file, line,__FUNCTION__);                \
-		fprintf(stderr, __VA_ARGS__);				\
-		fprintf(stderr, "\n");					\
-                fprintf(stderr, "### Test random seed was %i ###\033[0m\n",    \
-                        test_seed);                                     \
-                if (do_lock)                                            \
-                        TEST_LOCK();                                    \
-                test_curr->state = TEST_FAILED;                         \
-                if (fail_now && test_curr->mainfunc) {                  \
-                        tests_running_cnt--;                            \
-                        is_thrd = 1;                                    \
-                }                                                       \
-                if (do_lock)                                            \
-                        TEST_UNLOCK();                                  \
-		if (!fail_now) break;					\
-                if (test_assert_on_fail || !is_thrd)                    \
-                        assert(0);                                      \
-                else                                                    \
-                        thrd_exit(0);                                   \
-	} while (0)
+#define TEST_FAIL0(file,line,do_lock,fail_now,...)   \
+        test_fail0(__FILE__, __LINE__, __FUNCTION__, \
+                   do_lock, fail_now, __VA_ARGS__)
 
 /* Whine and abort test */
 #define TEST_FAIL(...) TEST_FAIL0(__FILE__,__LINE__,1,1,__VA_ARGS__)
@@ -175,6 +163,15 @@ struct test {
 /* Whine right away, mark the test as failed, but continue the test. */
 #define TEST_FAIL_LATER(...) TEST_FAIL0(__FILE__,__LINE__,1,0,__VA_ARGS__)
 
+/* Whine right away, maybe mark the test as failed, but continue the test. */
+#define TEST_FAIL_LATER0(LATER,...) TEST_FAIL0(__FILE__,__LINE__,1,!(LATER),__VA_ARGS__)
+
+#define TEST_FAILCNT()  (test_curr->failcnt)
+
+#define TEST_LATER_CHECK(...) do {                              \
+        if (test_curr->state == TEST_FAILED)                    \
+                TEST_FAIL("See previous errors. " __VA_ARGS__); \
+        } while (0)
 
 #define TEST_PERROR(call) do {						\
 		if (!(call))						\
@@ -220,12 +217,15 @@ struct test {
         } while (0)
 
 /* Skip the current test. Argument is textual reason (printf format) */
-#define TEST_SKIP(...) do {		     \
-		TEST_WARN("SKIPPING TEST: " __VA_ARGS__); \
-		TEST_LOCK();			     \
-		test_curr->state = TEST_SKIPPED;     \
-		TEST_UNLOCK();			     \
-	} while (0)
+#define TEST_SKIP(...) do {                                             \
+                TEST_WARN("SKIPPING TEST: " __VA_ARGS__);               \
+                TEST_LOCK();                                            \
+                test_curr->state = TEST_SKIPPED;                        \
+                if (!*test_curr->failstr)                               \
+                        rd_snprintf(test_curr->failstr,                 \
+                                    sizeof(test_curr->failstr), __VA_ARGS__); \
+                TEST_UNLOCK();                                          \
+        } while (0)
 
 
 void test_conf_init (rd_kafka_conf_t **conf, rd_kafka_topic_conf_t **topic_conf,
@@ -234,9 +234,6 @@ void test_conf_init (rd_kafka_conf_t **conf, rd_kafka_topic_conf_t **topic_conf,
 
 void test_wait_exit (int timeout);
 
-uint64_t test_id_generate (void);
-char *test_str_id_generate (char *dest, size_t dest_size);
-const char *test_str_id_generate_tmp (void);
 
 
 
@@ -294,6 +291,9 @@ typedef struct test_msgver_s {
 	int log_cnt;           /* Current number of warning logs */
 	int log_max;           /* Max warning logs before suppressing. */
 	int log_suppr_cnt;     /* Number of suppressed log messages. */
+
+        const char *msgid_hdr; /**< msgid string is in header by this name,
+                                * rather than in the payload (default). */
 } test_msgver_t;
 
 /* Message */
@@ -331,11 +331,19 @@ struct test_mv_vs {
         int64_t timestamp_max;
 
 	struct test_mv_mvec mvec;
+
+        /* Correct msgver for comparison */
+        test_msgver_t *corr;
 } vs;
 
 
 void test_msgver_init (test_msgver_t *mv, uint64_t testid);
 void test_msgver_clear (test_msgver_t *mv);
+int test_msgver_add_msg00 (const char *func, int line, test_msgver_t *mv,
+                           uint64_t testid,
+                           const char *topic, int32_t partition,
+                           int64_t offset, int64_t timestamp,
+                           rd_kafka_resp_err_t err, int msgnum);
 int test_msgver_add_msg0 (const char *func, int line,
 			  test_msgver_t *mv, rd_kafka_message_t *rkm);
 #define test_msgver_add_msg(mv,rkm) \
@@ -385,6 +393,12 @@ int test_msgver_verify0 (const char *func, int line, const char *what,
                                             .exp_cnt = expcnt})
 
 
+void test_msgver_verify_compare0 (const char *func, int line,
+                                  const char *what, test_msgver_t *mv,
+                                  test_msgver_t *corr, int flags);
+#define test_msgver_verify_compare(what,mv,corr,flags) \
+        test_msgver_verify_compare0(__FUNCTION__,__LINE__, what, mv, corr, flags)
+
 rd_kafka_t *test_create_handle (int mode, rd_kafka_conf_t *conf);
 
 /**
@@ -407,6 +421,8 @@ void test_produce_msgs (rd_kafka_t *rk, rd_kafka_topic_t *rkt,
                         uint64_t testid, int32_t partition,
                         int msg_base, int cnt,
 			const char *payload, size_t size);
+rd_kafka_resp_err_t test_produce_sync (rd_kafka_t *rk, rd_kafka_topic_t *rkt,
+                                       uint64_t testid, int32_t partition);
 
 rd_kafka_t *test_create_consumer (const char *group_id,
 				  void (*rebalance_cb) (
@@ -489,6 +505,8 @@ rd_kafka_resp_err_t test_auto_create_topic_rkt (rd_kafka_t *rk,
 rd_kafka_resp_err_t test_auto_create_topic (rd_kafka_t *rk, const char *name);
 int test_check_auto_create_topic (void);
 
+int test_get_partition_count (rd_kafka_t *rk, const char *topicname);
+
 int test_check_builtin (const char *feature);
 
 char *tsprintf (const char *fmt, ...) RD_FORMAT(printf, 1, 2);
@@ -507,5 +525,50 @@ void test_prepare_msg (uint64_t testid, int32_t partition, int msg_id,
 #if WITH_SOCKEM
 void test_socket_enable (rd_kafka_conf_t *conf);
 void test_socket_close_all (struct test *test, int reinit);
+int  test_socket_sockem_set_all (const char *key, int val);
 #endif
 
+void test_headers_dump (const char *what, int lvl,
+                        const rd_kafka_headers_t *hdrs);
+
+int32_t *test_get_broker_ids (rd_kafka_t *use_rk, size_t *cntp);
+
+void test_wait_metadata_update (rd_kafka_t *rk,
+                                rd_kafka_metadata_topic_t *topics,
+                                size_t topic_cnt,
+                                rd_kafka_metadata_topic_t *not_topics,
+                                size_t not_topic_cnt,
+                                int tmout);
+
+rd_kafka_event_t *
+test_wait_admin_result (rd_kafka_queue_t *q,
+                        rd_kafka_event_type_t evtype,
+                        int tmout);
+
+rd_kafka_resp_err_t
+test_wait_topic_admin_result (rd_kafka_queue_t *q,
+                              rd_kafka_event_type_t evtype,
+                              int tmout);
+
+rd_kafka_resp_err_t
+test_CreateTopics_simple (rd_kafka_t *rk,
+                          rd_kafka_queue_t *useq,
+                          char **topics, size_t topic_cnt,
+                          int num_partitions,
+                          void *opaque);
+rd_kafka_resp_err_t
+test_CreatePartitions_simple (rd_kafka_t *rk,
+                              rd_kafka_queue_t *useq,
+                              const char *topic,
+                              size_t total_part_cnt,
+                              void *opaque);
+
+rd_kafka_resp_err_t
+test_DeleteTopics_simple (rd_kafka_t *rk,
+                          rd_kafka_queue_t *useq,
+                          char **topics, size_t topic_cnt,
+                          void *opaque);
+
+rd_kafka_resp_err_t test_delete_all_test_topics (int timeout_ms);
+
+#endif /* _TEST_H_ */
